import { readdir, stat, lstat } from "node:fs/promises";
import { readdirSync, statSync, lstatSync } from "node:fs";
import { sep, resolve } from "node:path";
import picomatch from "picomatch";
const encoder = new TextEncoder();
const toUint8Array = encoder.encode.bind(encoder);
const decoder = new TextDecoder();
const toString = decoder.decode.bind(decoder);
const sepUint8Array = toUint8Array(sep);
const getEncoding = (dir) => dir instanceof Uint8Array ? "buffer" : "utf8";
const defaultOpts = {
  strict: false,
  stats: false,
  followSymlinks: false,
  exclude: void 0,
  include: void 0,
  insensitive: false
};
function makePath({ name }, dir, encoding) {
  if (encoding === "buffer") {
    return dir === "." ? name : Uint8Array.from([...dir, ...sepUint8Array, ...name]);
  } else {
    return dir === "." ? name : `${dir}${sep}${name}`;
  }
}
function build(dirent, path, stats, opts) {
  return {
    path,
    directory: (stats || dirent).isDirectory(),
    symlink: (stats || dirent).isSymbolicLink(),
    ...opts.stats ? { stats } : {}
  };
}
function makeMatchers({ include, exclude, insensitive }) {
  const opts = {
    dot: true,
    flags: insensitive ? "i" : void 0
  };
  return {
    includeMatcher: (include == null ? void 0 : include.length) ? (path) => picomatch(include, opts)(resolve(path)) : null,
    excludeMatcher: (exclude == null ? void 0 : exclude.length) ? (path) => picomatch(exclude, opts)(resolve(path)) : null
  };
}
async function* rrdir(dir, opts = {}, { includeMatcher, excludeMatcher, encoding } = {}) {
  if (includeMatcher === void 0) {
    opts = { ...defaultOpts, ...opts };
    ({ includeMatcher, excludeMatcher } = makeMatchers(opts));
    if (typeof dir === "string" && /[/\\]$/.test(dir))
      dir = dir.substring(0, dir.length - 1);
    encoding = getEncoding(dir);
  }
  let dirents = [];
  try {
    dirents = await readdir(dir, { encoding, withFileTypes: true });
  } catch (err) {
    if (opts.strict)
      throw err;
    yield { path: dir, err };
  }
  if (!dirents.length)
    return;
  for (const dirent of dirents) {
    const path = makePath(dirent, dir, encoding);
    if (excludeMatcher == null ? void 0 : excludeMatcher(encoding === "buffer" ? toString(path) : path))
      continue;
    const isSymbolicLink = opts.followSymlinks && dirent.isSymbolicLink();
    const encodedPath = encoding === "buffer" ? toString(path) : path;
    const isIncluded = !includeMatcher || includeMatcher(encodedPath);
    let stats;
    if (isIncluded) {
      if (opts.stats || isSymbolicLink) {
        try {
          stats = await (opts.followSymlinks ? stat : lstat)(path);
        } catch (err) {
          if (opts.strict)
            throw err;
          yield { path, err };
        }
      }
      yield build(dirent, path, stats, opts);
    }
    let recurse = false;
    if (isSymbolicLink) {
      if (!stats)
        try {
          stats = await stat(path);
        } catch {
        }
      if (stats && stats.isDirectory())
        recurse = true;
    } else if (dirent.isDirectory()) {
      recurse = true;
    }
    if (recurse)
      yield* rrdir(path, opts, { includeMatcher, excludeMatcher, encoding });
  }
}
async function rrdirAsync(dir, opts = {}, { includeMatcher, excludeMatcher, encoding } = {}) {
  if (includeMatcher === void 0) {
    opts = { ...defaultOpts, ...opts };
    ({ includeMatcher, excludeMatcher } = makeMatchers(opts));
    if (typeof dir === "string" && /[/\\]$/.test(dir))
      dir = dir.substring(0, dir.length - 1);
    encoding = getEncoding(dir);
  }
  const results = [];
  let dirents = [];
  try {
    dirents = await readdir(dir, { encoding, withFileTypes: true });
  } catch (err) {
    if (opts.strict)
      throw err;
    results.push({ path: dir, err });
  }
  if (!dirents.length)
    return results;
  await Promise.all(dirents.map(async (dirent) => {
    const path = makePath(dirent, dir, encoding);
    if (excludeMatcher == null ? void 0 : excludeMatcher(encoding === "buffer" ? toString(path) : path))
      return;
    const isSymbolicLink = opts.followSymlinks && dirent.isSymbolicLink();
    const encodedPath = encoding === "buffer" ? toString(path) : path;
    const isIncluded = !includeMatcher || includeMatcher(encodedPath);
    let stats;
    if (isIncluded) {
      if (opts.stats || isSymbolicLink) {
        try {
          stats = await (opts.followSymlinks ? stat : lstat)(path);
        } catch (err) {
          if (opts.strict)
            throw err;
          results.push({ path, err });
        }
      }
      results.push(build(dirent, path, stats, opts));
    }
    let recurse = false;
    if (isSymbolicLink) {
      if (!stats)
        try {
          stats = await stat(path);
        } catch {
        }
      if (stats && stats.isDirectory())
        recurse = true;
    } else if (dirent.isDirectory()) {
      recurse = true;
    }
    if (recurse)
      results.push(...await rrdirAsync(path, opts, { includeMatcher, excludeMatcher, encoding }));
  }));
  return results;
}
function rrdirSync(dir, opts = {}, { includeMatcher, excludeMatcher, encoding } = {}) {
  if (includeMatcher === void 0) {
    opts = { ...defaultOpts, ...opts };
    ({ includeMatcher, excludeMatcher } = makeMatchers(opts));
    if (typeof dir === "string" && /[/\\]$/.test(dir))
      dir = dir.substring(0, dir.length - 1);
    encoding = getEncoding(dir);
  }
  const results = [];
  let dirents = [];
  try {
    dirents = readdirSync(dir, { encoding, withFileTypes: true });
  } catch (err) {
    if (opts.strict)
      throw err;
    results.push({ path: dir, err });
  }
  if (!dirents.length)
    return results;
  for (const dirent of dirents) {
    const path = makePath(dirent, dir, encoding);
    if (excludeMatcher == null ? void 0 : excludeMatcher(encoding === "buffer" ? toString(path) : path))
      continue;
    const isSymbolicLink = opts.followSymlinks && dirent.isSymbolicLink();
    const encodedPath = encoding === "buffer" ? toString(path) : path;
    const isIncluded = !includeMatcher || includeMatcher(encodedPath);
    let stats;
    if (isIncluded) {
      if (opts.stats || isSymbolicLink) {
        try {
          stats = (opts.followSymlinks ? statSync : lstatSync)(path);
        } catch (err) {
          if (opts.strict)
            throw err;
          results.push({ path, err });
        }
      }
      results.push(build(dirent, path, stats, opts));
    }
    let recurse = false;
    if (isSymbolicLink) {
      if (!stats)
        try {
          stats = statSync(path);
        } catch {
        }
      if (stats && stats.isDirectory())
        recurse = true;
    } else if (dirent.isDirectory()) {
      recurse = true;
    }
    if (recurse)
      results.push(...rrdirSync(path, opts, { includeMatcher, excludeMatcher, encoding }));
  }
  return results;
}
export {
  rrdir,
  rrdirAsync,
  rrdirSync
};
